// This version requires region dependencies
#pragma oss assert("version.dependencies==regions")

unsigned long long ts;

#pragma oss task inout(a[start:end]) out(tmp[start:end])
void merge(double * const a, double * const tmp,
	   unsigned long long start, unsigned long long end)
{
	unsigned long long mid  = (start + end) / 2;
	unsigned long long i    = start;
	unsigned long long j    = mid + 1;
	unsigned long long h    = start;

	while ( (h <= mid) && (j <= end) )
	{
		if (a[h] <= a[j])
			tmp[i++] = a[h++];
		else
			tmp[i++] = a[j++];
	}

	if (h > mid) {
		for (unsigned long long k = j; k <= end; k++)
			tmp[i++] = a[k];
	}
	else {
		for (unsigned long long k = h; k <= mid; k++)
			tmp[i++] = a[k];
	}

	for (unsigned long long k = start; k <= end; k++) {
		a[k] = tmp[k];
	}
}

#pragma oss task weakinout(a[start:end]) weakout(tmp[start:end]) final(end-start+1 < ts)
void merge_sort_rec(double * const a, double * const tmp,
		    unsigned long long start, unsigned long long end)
{
	unsigned long long mid = (start + end) / 2;

	if (start < mid) {
		merge_sort_rec(a, tmp, start, mid);
	}

	if (mid+1 < end) {
		merge_sort_rec(a, tmp, mid + 1, end);
	}
	merge(a, tmp, start, end);
}


void merge_sort(double *a, unsigned long long n)
{
	double * const tmp =  new double[n];
	merge_sort_rec(a, tmp, 0, n-1);
	#pragma oss taskwait
	delete[] tmp;
}
